@@ -1,14 +1,20 @@
-use super::Location;
-use super::ParseResult;
-use super::Parser;
-use super::Trailing;
 #[cfg(test)]
 use super::*;
-use std::fmt::Debug;
+use std::convert::identity;
 
 #[test]
 fn test_one_of() {
     assert_succeed(one_of!(token("a"), token("b"), token("c")), "c", "c");
+    assert_succeed(
+        one_of!(
+            succeed!(|_a, _b| "a")
+                .keep(optional(token("a")))
+                .keep(token("x")),
+            token("b")
+        ),
+        "b",
+        "b",
+    );
     assert_fail(
         one_of!(token("a"), token("b"), token("c")),
         "",
@@ -36,7 +42,7 @@ fn test_update() {
                 state: Counter {
                     count: state.count + 1
                 },
-                bound: true,
+                committed: true,
             })
             .parse("a", Location { row: 1, col: 1 }, Counter { count: 0 }),
         ParseResult::Ok {
@@ -44,7 +50,7 @@ fn test_update() {
             location: Location { row: 1, col: 2 },
             output: "ab".to_string(),
             state: Counter { count: 1 },
-            bound: true,
+            committed: true,
         }
     );
     let counted_parser = token("a").update(|input, output, location, state: Counter| {
@@ -57,7 +63,7 @@ fn test_update() {
                 },
                 to: location,
                 state,
-                bound: false,
+                committed: false,
             }
         } else {
             ParseResult::Ok {
@@ -67,7 +73,7 @@ fn test_update() {
                 state: Counter {
                     count: state.count + 1,
                 },
-                bound: true,
+                committed: true,
             }
         }
     });
@@ -78,7 +84,7 @@ fn test_update() {
             location: Location { row: 1, col: 2 },
             output: "aa".to_string(),
             state: Counter { count: 1 },
-            bound: true,
+            committed: true,
         }
     );
     assert_eq!(
@@ -88,7 +94,7 @@ fn test_update() {
             from: Location { row: 1, col: 1 },
             to: Location { row: 1, col: 2 },
             state: Counter { count: 11 },
-            bound: false,
+            committed: false,
         }
     );
 }
@@ -379,50 +385,90 @@ fn test_newline_with_comment() {
 
 #[test]
 fn test_optional() {
+    assert_succeed(succeed!(identity).keep(optional(token("abc"))), "", None);
     assert_succeed(
-        succeed!(|a| a).keep(optional("yes", token("abc"))),
-        "",
-        "yes",
+        succeed!(identity)
+            .skip(token("abc"))
+            .keep(optional(token("hello"))),
+        "abc",
+        None,
+    );
+    assert_fail(
+        succeed!(|sign: Option<_>, n: f64| if let Some(_) = sign { -n } else { n })
+            .keep(optional(token("-")))
+            .keep(float()),
+        "null",
+        "I'm expecting a floating point number but found `n`.",
     );
     assert_succeed(
-        succeed!(|a| a)
+        succeed!(|sign: Option<_>, n: f64| if let Some(_) = sign { -n } else { n })
+            .keep(optional(token("-")))
+            .keep(float()),
+        "-3.2e2",
+        -3.2e2,
+    );
+}
+
+#[test]
+fn test_optional_with_default() {
+    assert_succeed(
+        succeed!(identity).keep(optional_with_default("default", token("abc"))),
+        "",
+        "default",
+    );
+    assert_succeed(
+        succeed!(identity)
             .skip(token("abc"))
-            .keep(optional("yes", token("hello"))),
+            .keep(optional_with_default("default", token("hello"))),
         "abc",
-        "yes",
+        "default",
+    );
+    assert_fail(
+        succeed!(|sign: bool, n: f64| if sign { -n } else { n })
+            .keep(optional_with_default(false, token("-").map(|_| true)))
+            .keep(float()),
+        "null",
+        "I'm expecting a floating point number but found `n`.",
+    );
+    assert_succeed(
+        succeed!(|sign: bool, n: f64| if sign { -n } else { n })
+            .keep(optional_with_default(false, token("-").map(|_| true)))
+            .keep(float()),
+        "-3.2e2",
+        -3.2e2,
     );
 }
 
 #[test]
 fn test_one_or_more_until() {
     assert_succeed(
-        succeed!(|list| list).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
         "ab",
         vec!["a"],
     );
     assert_succeed(
-        succeed!(|list| list).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
         "aaab",
         vec!["a", "a", "a"],
     );
     assert_fail(
-        succeed!(|list| list).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
         "b",
         "I'm expecting at least one occurrence of the intended string but reached the end delimiter."
     );
     assert_fail(
-        succeed!(|list| list).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
         "aaac",
         "I'm expecting either the intended string or the end delimiter. However, neither was found.",
     );
     assert_succeed(
-        succeed!(|list| list).keep(one_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(one_or_more_until(token("a"), token("b"))),
         "aaa",
         vec!["a", "a", "a"],
     );
 
     assert_succeed(
-        succeed!(|list| list).keep(one_or_more_until(
+        succeed!(identity).keep(one_or_more_until(
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
@@ -444,7 +490,7 @@ this is the 4th line
     );
 
     assert_succeed(
-        succeed!(|list| list).keep(one_or_more_until(
+        succeed!(identity).keep(one_or_more_until(
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
@@ -464,7 +510,7 @@ this is the 4th line
     );
 
     assert_fail(
-        succeed!(|list| list).keep(one_or_more_until(
+        succeed!(identity).keep(one_or_more_until(
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
@@ -478,38 +524,38 @@ this is the 4th line
 #[test]
 fn test_zero_or_more_until() {
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
         "ab",
         vec!["a"],
     );
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
         "aaab",
         vec!["a", "a", "a"],
     );
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
         "b",
         vec![],
     );
     assert_fail(
-        succeed!(|list| list).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
         "aaac",
         "I'm expecting either the intended string or the end delimiter. However, neither was found.",
     );
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
         "aaa",
         vec!["a", "a", "a"],
     );
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(token("a"), token("b"))),
+        succeed!(identity).keep(zero_or_more_until(token("a"), token("b"))),
         "",
         vec![],
     );
 
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(
+        succeed!(identity).keep(zero_or_more_until(
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
@@ -531,7 +577,7 @@ this is the 4th line
     );
 
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(
+        succeed!(identity).keep(zero_or_more_until(
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
@@ -551,7 +597,7 @@ this is the 4th line
     );
 
     assert_succeed(
-        succeed!(|list| list).keep(zero_or_more_until(
+        succeed!(identity).keep(zero_or_more_until(
             succeed!(|line| line)
                 .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
                 .skip(token("\n")),
@@ -560,37 +606,177 @@ this is the 4th line
         "<end>",
         vec![],
     );
+
+    // edge case: the language of parser may be a superset of end_parser
+    assert_succeed(
+        succeed!(identity).keep(zero_or_more_until(
+            succeed!(|line| line)
+                .keep(take_chomped(chomp_while0c(|c| *c != '\n', "line")))
+                .skip(token("\n")),
+            token("<end>"),
+        )),
+        "<end>\n",
+        vec![],
+    );
+}
+
+#[test]
+fn test_any_char() {
+    assert_eq!(
+        any_char("a character").parse("a", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 2 },
+            output: 'a',
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_char("a character").parse("\n", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 2, col: 1 },
+            output: '\n',
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_char("a character").parse("혇", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 3 },
+            output: '혇',
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_char("a character").parse("我", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 3 },
+            output: "我",
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_char("a character").parse("🤣", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 3 },
+            output: '🤣',
+            state: (),
+            committed: false,
+        }
+    );
+}
+
+#[test]
+fn test_any_grapheme() {
+    assert_eq!(
+        any_grapheme("a grapheme").parse("a", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 2 },
+            output: "a",
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_grapheme("a grapheme").parse("\n", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 2, col: 1 },
+            output: "\n",
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_grapheme("a grapheme").parse("\r\n", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 2, col: 1 },
+            output: "\r\n",
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_grapheme("a grapheme").parse("혇", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 3 },
+            output: "혇",
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_grapheme("a grapheme").parse("我", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 3 },
+            output: "我",
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_grapheme("a grapheme").parse("🤣", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 3 },
+            output: "🤣",
+            state: (),
+            committed: false,
+        }
+    );
+    assert_eq!(
+        any_grapheme("a grapheme").parse("👩‍🔬", Location { row: 1, col: 1 }, ()),
+        ParseResult::Ok {
+            input: "",
+            location: Location { row: 1, col: 3 },
+            output: "👩‍🔬",
+            state: (),
+            committed: false,
+        }
+    );
 }
 
 #[test]
 fn test_chomp_if() {
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_if(|c| c == "👨‍👩‍👧‍👦", "family emoji"))),
+        succeed!(identity).keep(take_chomped(chomp_if(|c| c == "👨‍👩‍👧‍👦", "family emoji"))),
         "👨‍👩‍👧‍👦",
         "👨‍👩‍👧‍👦".to_string(),
     );
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_if(|c| c == "혇", "Korean"))),
+        succeed!(identity).keep(take_chomped(chomp_if(|c| c == "혇", "Korean"))),
         "혇",
         "혇".to_string(),
     );
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_if(|c| c == "齉", "Chinese"))),
+        succeed!(identity).keep(take_chomped(chomp_if(|c| c == "齉", "Chinese"))),
         "齉",
         "齉".to_string(),
     );
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_if(|c| c == "ë", "IPA Symbol"))),
+        succeed!(identity).keep(take_chomped(chomp_if(|c| c == "ë", "IPA Symbol"))),
         "ë",
         "ë".to_string(),
     );
     assert_fail(
-        succeed!(|s| s).keep(chomp_if(|c| c == "a", "a letter `a`")),
+        succeed!(identity).keep(chomp_if(|c| c == "a", "a letter `a`")),
         "",
         "I'm expecting a letter `a` but reached the end of input.",
     );
     assert_fail(
-        succeed!(|s| s).keep(chomp_if(|_| true, "any character")),
+        succeed!(identity).keep(chomp_if(|_| true, "any character")),
         "",
         "I'm expecting any character but reached the end of input.",
     );
@@ -599,12 +785,12 @@ fn test_chomp_if() {
 #[test]
 fn test_chomp_ifc() {
     assert_fail(
-        succeed!(|s| s).keep(chomp_ifc(|c| *c == 'a', "a letter `a`")),
+        succeed!(identity).keep(chomp_ifc(|c| *c == 'a', "a letter `a`")),
         "",
         "I'm expecting a letter `a` but reached the end of input.",
     );
     assert_fail(
-        succeed!(|s| s).keep(chomp_ifc(|_| true, "any character")),
+        succeed!(identity).keep(chomp_ifc(|_| true, "any character")),
         "",
         "I'm expecting any character but reached the end of input.",
     );
@@ -613,17 +799,17 @@ fn test_chomp_ifc() {
 #[test]
 fn test_chomp_while1() {
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_while1(|c| c == "我", "character '我'"))),
+        succeed!(identity).keep(take_chomped(chomp_while1(|c| c == "我", "character '我'"))),
         "我我我你",
         "我我我".to_string(),
     );
     assert_fail(
-        succeed!(|s| s).keep(take_chomped(chomp_while1(|c| c == "我", "character `我`"))),
+        succeed!(identity).keep(take_chomped(chomp_while1(|c| c == "我", "character `我`"))),
         "你",
         "I'm expecting character `我` but found `你`.",
     );
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_while1(
+        succeed!(identity).keep(take_chomped(chomp_while1(
             |c| c != "👩‍👩‍👦‍👦",
             "any emoji except the family emoji",
         ))),
@@ -631,7 +817,7 @@ fn test_chomp_while1() {
         "🥲🥰🏉".to_string(),
     );
     assert_fail(
-        succeed!(|s| s).keep(take_chomped(chomp_while1(
+        succeed!(identity).keep(take_chomped(chomp_while1(
             |c| c != "👩‍👩‍👦‍👦",
             "any emoji except the family emoji",
         ))),
@@ -643,17 +829,17 @@ fn test_chomp_while1() {
 #[test]
 fn test_chomp_while0() {
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_while0(|c| c == "我", "character '我'"))),
+        succeed!(identity).keep(take_chomped(chomp_while0(|c| c == "我", "character '我'"))),
         "我我我你",
         "我我我".to_string(),
     );
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_while0(|c| c == "我", "character '我'"))),
+        succeed!(identity).keep(take_chomped(chomp_while0(|c| c == "我", "character '我'"))),
         "你",
         "".to_string(),
     );
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_while0(
+        succeed!(identity).keep(take_chomped(chomp_while0(
             |c| c != "👩‍👩‍👦‍👦",
             "any emoji except the family emoji",
         ))),
@@ -661,7 +847,7 @@ fn test_chomp_while0() {
         "🥲🥰🏉".to_string(),
     );
     assert_succeed(
-        succeed!(|s| s).keep(take_chomped(chomp_while0(
+        succeed!(identity).keep(take_chomped(chomp_while0(
             |c| c != "👩‍👩‍👦‍👦",
             "any emoji except the family emoji",
         ))),
@@ -669,3 +855,17 @@ fn test_chomp_while0() {
         "".to_string(),
     );
 }
+
+#[test]
+fn test_digits() {
+    assert_succeed(
+        digits("a nonnegative int without leading zeroes", false),
+        "1028923",
+        "1028923".to_string(),
+    );
+    assert_fail(
+        digits("a nonnegative int without leading zeroes", false),
+        "a",
+        "I'm expecting a nonnegative int without leading zeroes but found `a`.",
+    );
+}
